import numpy as np

from math import exp
from typing import Any
from typing import Tuple
from typing import Literal
from typing import Sequence
from typing import Generator
from numpy.typing import NDArray
from pandas.core.series import Series

from scipy import stats
from scipy.stats import f
from scipy.stats import levene
from scipy.stats import ks_1samp
from scipy.stats import f_oneway
from scipy.stats import anderson
from scipy.stats import skewtest
from scipy.stats import ttest_ind
from scipy.stats import ttest_1samp
from scipy.stats import kurtosistest
from scipy.stats import fisher_exact
from scipy.stats import mannwhitneyu
from scipy.stats._distn_infrastructure import rv_continuous

from statsmodels.stats.proportion import test_proportions_2indep

from .._typing import NumericSample1D


__all__ = [
    'chunker',
    'ensure_generic',
    'anderson_darling_test',
    'all_normal',
    'kolmogorov_smirnov_test',
    'f_test',
    't_test',
    'levene_test',
    'variance_stability_test',
    'mean_stability_test',
    'position_test',
    'variance_test',
    'proportions_test',
    'kurtosis_test',
    'skew_test',]

# TDOD: further tests:
# from scipy.stats import chi2
# from scipy.stats import ansari
# from scipy.stats import ranksums
# from scipy.stats import wilcoxon
# from statsmodels.stats.proportion import proportion_confint
# from statsmodels.stats.proportion import confint_proportions_2indep

def chunker(
        samples: Sequence[Any] | Series | NDArray,
        n_sections: int
        ) -> Generator[NDArray, Any, None]:
    """Divides the data into a specified number of sections.
    
    Parameters
    ----------
    sample : Sequence[Any]
        A one-dimensional array-like object containing the samples.
    n_sections : int
        Amount of sections to divide the data into.
        
    Yields
    ------
    NDArray
        A section of the data.
    
    Notes
    -----
    If equal-sized sections cannot be created, the first sections are 
    one larger than the rest.

    If more sections are to be created than the number of samples, 
    empty arrays are created.
    """
    assert n_sections > 0 and isinstance(n_sections, int)
    size, extras = divmod(len(samples), n_sections)
    sizes = extras*[size + 1] + (n_sections - extras)*[size]
    slicing_positions = np.array([0] + sizes).cumsum()

    _samples = np.asarray(samples)
    for i in range(n_sections):
        yield _samples[slicing_positions[i]:slicing_positions[i+1]]

def ensure_generic(
        dist: str | rv_continuous
        ) -> rv_continuous:
    """If the input is a string representing a distribution, convert it
    to a rv_continuous object.
    
    Parameters
    ----------
    dist : str or rv_continuous
        The distribution to convert. Can be either a string representing
        a distribution or a rv_continuous object.
    
    Returns
    -------
    rv_continuous
        The converted rv_continuous object if the input is a
        string representing a distribution, otherwise returns the input
        distribution directly.
    
    Raises
    ------
    AssertionError
        If the input is a string that does not represent a valid 
        distribution in scipy.stats.
    """
    _dist: rv_continuous
    _dist = getattr(stats, dist) if isinstance(dist, str) else dist
    assert hasattr(stats, getattr(_dist, 'name')), (
        f'{dist} is not a valid distribution')
    return _dist

def anderson_darling_test(
        sample: NumericSample1D
        ) -> Tuple[float, float]:
    """The Anderson-Darling test compares the measured values with the 
    theoretical values of a given distribution (in this case the normal 
    distribution). This test is considered to be one of the most 
    powerful tests for normal distribution for both small and large 
    sample sizes.

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.

    Returns
    -------
    p : float
        The p-value for the test
    A_star : float
        The adjusted Anderson Darling test statistic
    
    Notes
    -----
    This test was inspired by the Excel Addin by Charles Zaiontz, see:
    https://real-statistics.com/non-parametric-tests/goodness-of-fit-tests/anderson-darling-test/
    """
    N = len(sample)
    A2: float = anderson(sample, dist='norm')[0] # type: ignore
    A_star = A2*(1 + 0.75/N + 2.25/N**2)
    if 13 <= A_star:
        p = 0.0
    elif 0.6 <= A_star < 13:
        p = exp(1.2937 - 5.709*A_star + 0.0186*(A_star**2))
    elif 0.34 < A_star < 0.6:
        p = exp(0.9177 - 4.279*A_star - 1.38*(A_star**2))
    elif 0.2 < A_star <= 0.34:
        p = 1 - exp(42.796*A_star - 59.938*(A_star**2) - 8.318)
    else:
        p = 1 - exp(101.14*A_star - 223.73*(A_star**2) - 13.436)
    return p, A_star

def all_normal(
        *samples: NumericSample1D,
        p_threshold: float = 0.05
        ) -> bool:
    """Performs the Anderson-Darling test against the normal
    distribution for each given sample data. Only one-dimensional
    samples are accepted.
    
    Parameters
    ----------
    *samples : NumericSample1D
        One or more one-dimensional array-like objects containing the
        samples.
    p_threshold : float, optional
        The threshold p-value for significance (default is 0.05).
    
    Returns
    -------
    bool
        True if all p-values are greater than the specified p_threshold,
        False otherwise.
    
    Raises
    ------
    AssertionError
        If p_threshold is not within the range (0, 1).
    """
    assert 0 < p_threshold < 1, 'p_threshold must be within (0, 1)'
    return all([anderson_darling_test(x)[0] > p_threshold for x in samples])

def kolmogorov_smirnov_test(
        sample: NumericSample1D,
        dist: str | rv_continuous,
        alternative: Literal['two-sided', 'less', 'greater'] = 'two-sided'
        ) -> Tuple[float, float, Tuple[float, ...]]:
    """Perform a one-sample Kolmogorov-Smirnov-Test. This hypothesis
    test compares the underlying distribution F(x) of a sample against a 
    given distribution G(x). This test is valid only for continuous 
    distributions.
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    dist : str or scipy.stats rv_continous
        If a string, it should be the name of a continous distribution 
        in scipy.stats, which will be used as the cdf function.
    alternative : {'two-sided', 'less', 'greater'}, optional
        The alternative hypothesis to use for the test.
        'two-sided' : the cdf of the distribution is not the same as the 
            cdf of the sample
        'less' : the cdf of the distribution is the same as or less than 
            the cdf of the sample
        'greater' : the cdf of the distribution is the same as or greater 
            than the cdf of the sample
    
    Returns
    -------
    p : float
        The two-tailed p-value for the test
    D : float
        Kolmogorov-Smirnov test statistic, either D, D+ or D-.
    params : Tuple[float, ...]
        Estimates for any shape parameters (if applicable), followed by 
        those for location and scale. For most random variables, shape 
        statistics will be returned, but there are exceptions 
        (e.g. ``norm``).
    """
    dist = ensure_generic(dist)
    params = dist.fit(sample)
    D, p = ks_1samp(
        sample, cdf=dist.cdf, args=params, alternative=alternative)
    return p, D, params # type: ignore

def f_test(
        sample1: NumericSample1D,
        sample2: NumericSample1D
        ) -> Tuple[float, float]:
    """The F-test is a test for equal variances between two populations. 
    The probability distribution on which the F-test is based is called 
    the F-distribution (also Fisher distribution). 

    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second
        sample.

    Returns
    -------
    p : float
        The p-value for the test
    F : float
        The f-test statistic
    """
    F = float(np.var(sample1, ddof=1) / np.var(sample2, ddof=1))
    dof1, dof2 = len(sample1)-1, len(sample2)-1
    cumulated = float(f.cdf(F, dof1, dof2))
    p = 2 * min(cumulated, 1-cumulated)
    return p, F

def t_test(
        sample: NumericSample1D,
        mu: float = 0,
        alternative: Literal['two-sided', 'less', 'greater'] = 'two-sided'
        ) -> Tuple[float, float, int]:
    """Perform one sample t-test.
    
    The t-test tests the null hypothesis that the mean of a sample is
    equal to a given population mean.
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the sample.
    mu : float, optional
        The hypothesized mean of the population, by default 0
    alternative : {'two-sided', 'less', 'greater'}, optional
        The alternative hypothesis to use for the test.
        'two-sided' : the mean of the population is not equal to mu
        'less' : the mean of the population is less than mu
        'greater' : the mean of the population is greater than mu
    
    Returns
    -------
    p : float
        The p-value for the test
    t : float
        The t-test statistic
    df : int
        The degrees of freedom
    """
    result = ttest_1samp(sample, mu, alternative=alternative)
    p, t, df = float(result.pvalue), float(result.statistic), int(result.df)  # type: ignore
    return p, t, df

def levene_test(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        heavy_tailed: bool = False
        ) -> Tuple[float, float]:
    """Perform Levene test for equal variances.
    The Levene test tests the null hypothesis that all input samples are 
    from populations with equal variances.
    
    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second
        sample.
    heavy_tailed : bool
        set True if data is heavy tailed, by default False 
    
    Returns
    -------
    p : float
        p-value for the test
    L : float
        Levene test statistic
    """
    center = 'trimmed' if heavy_tailed else 'median'
    L, p = levene(sample1, sample2, center=center)
    return p, L

def variance_stability_test(
        sample: NumericSample1D,
        n_sections: int = 3
        ) -> Tuple[float, float]:
    """Perform Levene test for equal variances within one sample.
    
    Divides the data into the number of n_sections. A Levene test is 
    then performed between these intercepts to check whether the 
    variance remains stable
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    n_sections : int, optional
        Amount of sections to divide the data into, by default 3

    Returns
    -------
    p : float
        p-value for the test
    L : float
        Levene test statistic
    """
    L, p = levene(*chunker(sample, n_sections), center='median')
    return p, L

def mean_stability_test(
        sample: NumericSample1D,
        n_sections: int = 3
        ) -> Tuple[float, float]:
    """Perform one-way ANOVA for equal means within one sample.
    
    Divides the data into the number of n_sections. A f_oneway test is 
    then performed between these intercepts to check whether the 
    mean remains stable
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    n_sections : int, optional
        Amount of sections to divide the data into, by default 3

    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        The computed F statistic of the test.
    """
    statistic, p = f_oneway(*chunker(sample, n_sections))
    return p, statistic

def position_test(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        equal_var: bool = True,
        normal: bool | None = None,
        u_test: bool=True
        ) -> Tuple[float, float, str]:
    """calculate the test for the means of *two independent* samples of 
    scores.
    This is a two-sided test for the null hypothesis that 2 independent
    samples have identical average (expected) values. This test assumes
    that the populations have identical variances by default.
    If u_test is true and normal is false perform the Mann-Whitney U 
    rank test on two independent samples.
    The Mann-Whitney U test is a nonparametric test of the null 
    hypothesis that the distribution underlying sample x is the same as 
    the distribution underlying sample y. It is often used as a test of 
    difference in location between distributions.


    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second
        sample.
    equal_var : bool, optional
        If True (default), perform a standard independent 2 sample test 
        that assumes equal population variances. If False, perform 
        Welch's t-test, which does not assume equal population variance
    normal : bool or None
        Set to True if both sample data are normally distributed. If 
        True, perform a t-test. If False and u_test is True, perform a 
        Mann Whitney U test. If None, an Anderson-Darling test for 
        normal distribution is performed for both sample data. If one of
        the two data sets is not normally distributed, normal is set to 
        False, by default None
    u_test : bool, optional
        If True and data are not normally distributed, perform a Mann 
        Whitney U test, by default True

    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        f if normal, else Levene test statistic
    test : string
        name of performed test
    """
    if not isinstance(normal, bool):
        normal = all_normal(sample1, sample2)

    if u_test and not normal:
        statistic, p = mannwhitneyu(
            sample1, sample2, alternative='two-sided', method='asymptotic')
        test = 'Mann-Whitney-U'
    else:
        statistic, p = ttest_ind(sample1, sample2, equal_var=equal_var)
        test = 't'
    return p, statistic, test # type: ignore

def variance_test(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        normal: bool | None = None,
        heavy_tailed: bool = False
        ) -> Tuple[float, float, str]:
    """Perform test for equal variances of two independent variables.
    This test tests the null hypothesis that all input samples are 
    from populations with equal variances.
    
    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second
        sample.
    normal : bool or None
        Set to True if both sample data are normally distributed. If 
        true, an F-test is performed, otherwise a Levene test. If None, 
        an Anderson-Darling test for normal distribution is performed 
        for both sample data. If one of the two data sets is not 
        normally distributed, normal is set to False, by default None
    heavy_tailed : bool
        set True if data is heavy tailed. Is only taken into account if 
        normal is False , by default False
    
    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        f if normal, else Levene test statistic
    test : string
        name of performed test
    """
    if not isinstance(normal, bool):
        normal = all_normal(sample1, sample2)
    
    if normal:
        p, statistic = f_test(sample1, sample2)
        test = 'F'
    else:
        p, statistic = levene_test(sample1 ,sample2, heavy_tailed)
        test = 'Levenes'
        # statistic, p = ansari(sample1, sample2, alternative='two-sided')
    return p, statistic, test

def proportions_test(
        events1: int,
        observations1: int,
        events2: int,
        observations2: int,
        decision_threshold: int = 1000
        ) -> Tuple[float, float, str]:
    """Hypothesis test for comparing two independent proportions
    This assumes that we have two independent binomial samples.
    
    Fisher's exact test is one of exact tests. Especially when more than 
    20% of cells have expected frequencies < 5, we need to use Fisher's 
    exact test because applying approximation method is inadequate. 
    Fisher's exact test assesses the null hypothesis of independence 
    applying hypergeometric distribution of the numbers in the cells 
    of the table. 

    Parameters
    ----------
    events1 : int
        Counted number of events of sample 1.
    observations1 : int
        Total number of observations of sample 1.
    events2 : int
        counted number of events of sample 2.
    observations2 : int
        Total number of observations of sample 2.
    decision_threshold : int, optional
        if the sum of sample size (observations1 + observations2) is greater
        than decision_threshold, the Fisher exact test is performed, 
        by default 1000

    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        test statistic
    test : string
        name of performed test
    """
    test = ''
    if observations1 + observations2 > decision_threshold:
        table = np.array([[events1, observations1], [events2, observations2]])
        statistic, p = fisher_exact(table, alternative='two-sided')
        test = 'Exakter Fisher'
    else:
        statistic, p = test_proportions_2indep(
            events1, observations1, events2, observations2, 
            method='wald', alternative='two-sided', return_results=False)
        test = 'Wald'
    return p, statistic, test # type: ignore

def kurtosis_test(
        sample: NumericSample1D
        ) -> Tuple[float, float]:
    """Two sided hypothesis test whether a dataset has normal kurtosis.

    This function tests the null hypothesis that the kurtosis of the 
    population from which the sample was drawn is that of the normal 
    distribution. Performs the calculations ignoring nan values
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.

    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        The computed z-score for this test
    """
    statistic, p = kurtosistest(
        sample, nan_policy='omit', alternative='two-sided')
    return p, statistic

def skew_test(
        sample: NumericSample1D
        ) -> Tuple[float, float]:
    """Two sided hypothesis whether the skew is different from the 
    normal distribution.

    This function tests the null hypothesis that the skewness of the 
    population that the sample was drawn from is the same as that of a 
    corresponding normal distribution. Performs the calculations 
    ignoring nan values.
    
    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.

    Returns
    -------
    p : float
        p-value for the test
    statistic : float
        The computed z-score for this test"""
    statistic, p = skewtest(sample, nan_policy='omit', alternative='two-sided')
    return p, statistic



# source for ci: https://aegis4048.github.io/comprehensive_confidence_intervals_for_python_developers#conf_int_of_var
import numpy as np
import pandas as pd

from math import sqrt
from typing import List
from typing import Tuple
from typing import Callable
from scipy.stats import t
from scipy.stats import f
from scipy.stats import chi2
from scipy.stats import norm
from numpy.typing import NDArray
from pandas.core.frame import DataFrame
from statsmodels.stats.proportion import proportion_confint
from statsmodels.stats.proportion import confint_proportions_2indep
from statsmodels.regression.linear_model import RegressionResults
from statsmodels.stats.outliers_influence import OLSInfluence
from statsmodels.sandbox.regression.predstd import wls_prediction_std

from .._typing import NumericSample1D


__all__ = [
    'sem',
    'mean_ci',
    'median_ci',
    'variance_ci',
    'stdev_ci',
    'proportion_ci',
    'cp_ci',
    'cpk_ci',
    'bonferroni_ci',
    'delta_mean_ci',
    'delta_variance_ci',
    'delta_stdev_ci',
    'delta_proportions_ci',
    'fit_ci',
    'prediction_ci',
    'confidence_to_alpha',]


def sem(
        sample: NumericSample1D,
        ddof: int = 1) -> float:
    """Calculate the standard error of the mean (or standard error of
    measurement) of the values in the input array.

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    ddof : int, optional
        Delta degrees-of-freedom. How many degrees of freedom to adjust
        for bias in limited sample relative to the population estimate
        of variance, by default 1.

    Returns
    -------
    se : float
        The standard error of the mean in the sample.
    """
    se = np.std(sample, ddof=ddof) / np.sqrt(np.size(sample))
    return se

def mean_ci(
        sample: NumericSample1D,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Two sided confidence interval for mean of data.

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    level : float in (0, 1), optional
        confidence level, by default 0.95
    n_groups : int, optional
        Used for Bonferroni method. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1

    Returns
    -------
    x_bar : float
        expected value
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    
    Notes
    -----
    The underlying `t.interval` function assumes that the data follows a
    t-distribution. Additionally, this method assumes that the sample is 
    representative of the population and that the data is independent 
    and identically distributed.
    """ 
    level = 1 - confidence_to_alpha(level, two_sided=False, n_groups=n_groups)
    se = sem(sample)
    dof = len(sample)-1
    x_bar = float(np.mean(sample))
    lower, upper = t.interval(level, dof, loc=x_bar, scale=se)
    return x_bar, lower, upper

def median_ci(
        sample: NumericSample1D,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Two sided confidence interval for median of data

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    level : float in (0, 1), optional
        confidence level, by default 0.95
    n_groups : int, optional
        Used for Bonferroni method. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1

    Returns
    -------
    median : float
        median of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    
    Notes
    -----
    The underlying `t.interval` function assumes that the data follows a
    t-distribution. Additionally, this method assumes that the sample is 
    representative of the population and that the data is independent 
    and identically distributed.
    """
    level = 1 - confidence_to_alpha(level, two_sided=False, n_groups=n_groups)  # adjust ci if Bonferroni method is performed
    se = sem(sample)
    dof = len(sample) - 1
    median = float(np.median(sample))
    lower, upper = t.interval(level, dof, loc=median, scale=se)
    return median, lower, upper

def variance_ci(
        sample: NumericSample1D,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Two sided confidence interval for variance of data

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    level : float in (0, 1), optional
        confidence level, by default 0.95
    n_groups : int, optional
        Used for Bonferroni method. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1

    Returns
    -------
    s2 : float
        variance of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    """
    alpha = confidence_to_alpha(level, two_sided=True, n_groups=n_groups)
    dof = len(sample) - 1
    s2 = float(np.var(sample, ddof=1)) # do not remove ddof=1, default is 0!
    upper = float(dof * s2 / chi2.ppf(alpha, dof))
    lower = float(dof * s2 / chi2.ppf(1 - alpha, dof))
    return s2, lower, upper

def stdev_ci(
        sample: NumericSample1D,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Two sided confidence interval for standard deviation of data

    Parameters
    ----------
    sample : NumericSample1D
        A one-dimensional array-like object containing the samples.
    level : float in (0, 1), optional
        confidence level, by default 0.95
    n_groups : int, optional
        Used for Bonferroni method. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1

    Returns
    -------
    s : float
        variance of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    """
    s, lower, upper = tuple(map(sqrt, variance_ci(sample, level, n_groups)))
    return s, lower, upper

def proportion_ci(
        events: int,
        observations: int,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Confidence interval for a binomial proportion with a asymptotic 
    normal approximation.

    Parameters
    ----------
    events : int
        Counted number of events.
    observations : int
        Total number of observations.
    level : float in (0, 1), optional
        Confidence level, default 0.95
    n_groups : int, optional
        Used for Bonferroni method. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1

    Returns
    -------
    portion : float
        Portion as ratio events/observations.
    lower, upper : float
        The lower and upper confidence level with coverage approximately 
        ci. 
    """
    alpha = confidence_to_alpha(level, two_sided=False, n_groups=n_groups)
    lower, upper = proportion_confint(events, observations, alpha, 'normal')
    portion = events/observations
    return portion, lower, upper # type: ignore

def cp_ci(
        cp: float,
        n_samples: int,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Calculate the confidence interval for the Cp estimator of process 
    capability.

    This function computes the confidence interval for the Cp value, 
    which quantifies the capability of a process to produce output 
    within specified limits. The width of the confidence interval is 
    influenced by the number of samples and the specified confidence 
    level. The calculation is based on the Chi-squared (χ²) distribution.

    Parameters
    ----------
    cp : float
        The Cp value representing the process capability.
    n_samples : int
        The total number of samples used in the estimation.
    level : float, optional
        The desired confidence level for the interval, expressed as a 
        decimal. Default is 0.95 (95% confidence).
    n_groups : int, optional
        The number of groups for Bonferroni correction to adjust for 
        multiple comparisons. Default is 1, indicating no correction.

    Returns
    -------
    cp : float
        For coherence with the other functions, the Cp value is 
        returned unchanged.
    lower, upper : float
        The lower and upper bounds of the confidence interval for the Cp 
        value.

    Notes
    -----
    - The confidence interval provides a range within which the true 
      process capability is expected to lie, given the sample data.
    - Ensure that the number of samples is greater than the number of 
      groups to avoid invalid degrees of freedom.
    
    References
    ----------
    For more information on Cpk confidence intervals, visit:
    https://www.qimacros.com/process-capability-analysis/cpk-confidence-intervals/
    """
    alpha = confidence_to_alpha(level, two_sided=False, n_groups=n_groups)
    dof = n_samples - n_groups
    lower = cp * float(chi2.ppf(alpha, dof) / (dof))**0.5
    upper = cp * float(chi2.ppf(1 - alpha, dof) / (dof))**0.5
    return cp, lower, upper

def cpk_ci(
        cpk: float,
        n_samples: int,
        level: float = 0.95,
        n_groups: int = 1
        ) -> Tuple[float, float, float]:
    """Calculate the confidence interval for the Cpk estimator of 
    process capability.

    This function computes the confidence interval for the Cpk value, 
    which assesses the capability of a process to meet both upper and 
    lower specification limits. The confidence interval is influenced by 
    the number of samples, the specified confidence level, and the 
    number of groups for Bonferroni correction. The calculation 
    incorporates the normal distribution.

    Parameters
    ----------
    cpk : float
        The Cpk value representing the process capability, accounting 
        for both process mean and variability.
    n_samples : int
        The total number of samples used in the estimation.
    level : float, optional
        The desired confidence level for the interval, expressed as a 
        decimal. Default is 0.95 (95% confidence).
    n_groups : int, optional
        The number of groups for Bonferroni correction to adjust for 
        multiple comparisons. Default is 1, indicating no correction.

    Returns
    -------
    cpk : float
        For coherence with the other functions, the Cpk value is 
        returned unchanged.
    lower, upper : float
        The lower and upper bounds of the confidence interval for the 
        Cpk value.

    Notes
    -----
    - The confidence interval provides a range within which the true 
      process capability is expected to lie, given the sample data.
    - Ensure that the number of samples is greater than the number of 
      groups to avoid invalid degrees of freedom.
    - The whisker term in the calculation accounts for variability in 
    the Cpk estimator based on sample size and distribution.
    
    References
    ----------
    For more information on Cpk confidence intervals, visit:
    https://www.qimacros.com/process-capability-analysis/cpk-confidence-intervals/
    """
    alpha = confidence_to_alpha(level, two_sided=False, n_groups=n_groups)
    dof = n_samples - n_groups
    crit = norm.ppf(1-alpha)
    confidence = cpk * crit * np.sqrt(1/(9*n_samples*cpk) + 1/(2*dof))
    lower = cpk - confidence
    upper = cpk + confidence
    return cpk, lower, upper

def bonferroni_ci(
        data: DataFrame,
        target: str,
        feature: str | List[str],
        level: float = 0.95,
        ci_func: Callable = stdev_ci,
        n_groups: int | None = None,
        name: str='midpoint'
        ) -> DataFrame:
    """Calculate confidence interval after bonferroni correction.
    The Bonferroni correction is a method to adjust the significance 
    level alpha.

    Parameters
    ----------
    data : DataFrame
        data frame containing sample and feature data
    target : str
        name of target sample data column
    feature : str | List[str]
        name of categorical feature. The confidence intervals are 
        calculated separately for these groups
    level : float in (0, 1), optional
        confidence level, default 0.95
    ci_func : {mean_ci, stdev_ci, variance_ci}
        function to calculate needed confidence interval that returns
        the values in order: midpoint, lower ci, upper ci
    n_groups : int, optional
        Used for Bonferroni correction. 
        Amount of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, 
        If none is given, it calculates the number based on the given 
        groups (ngroups attribute of groupby object), by default None
    name : str, optional
        name of midpoints, by default 'midpoint'

    Returns
    -------
    data : DataFrame
        data containing groups, midpoints and confidence limits
    
    Notes
    -----
    The Bonferroni correction is always necessary if you carry out 
    several "multiple" tests. In this case, the probability of the type
    I error for all tests together is no longer 5% (or 1%), but
    significantly more. This means that the risk that you will receive
    at least one significant result, even though there is no effect at
    all, is significantly increased with multiple tests. This is also
    referred to as alpha error accumulation or alpha inflation.
    """
    columns = [name, 'lower', 'upper']
    groups = data.groupby(feature)[target]
    n_groups = n_groups if isinstance(n_groups, int) else groups.ngroups
    data = pd.DataFrame(
            groups.agg(lambda x: ci_func(x, level, n_groups)).to_list(),
            columns = columns)
    
    if isinstance(feature, str):
        cat_values = list(groups.indices.keys())
    else:
        cat_values = list(map(list, groups.indices.keys())) # type: ignore
    
    data[feature] = cat_values
    return data

def delta_mean_ci(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        level: float = 0.95
        ) -> Tuple[float, float, float]:
    """Two sided confidence interval for mean difference of two
    independent variables.

    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first
        samples.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second
        samples.
    level : float in (0, 1), optional
        confidence level between 0 and 1, by default 0.95

    Returns
    -------
    delta : float
        Difference of means of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    """
    alpha = confidence_to_alpha(level, two_sided=True)
    n1 = len(sample1)
    n2 = len(sample2)
    s1 = np.var(sample1, ddof=1)
    s2 = np.var(sample2, ddof=1)
    # pooled standard deviation:
    s = np.sqrt(((n1 - 1) * s1 + (n2 - 1) * s2) / (n1 + n2 - 2))
    dof = n1 + n2 - 2
    t_crit = t.ppf(1 - alpha, dof)
    delta = float(np.mean(sample1) - np.mean(sample2))
    confidence = s*t_crit*np.sqrt(1/len(sample1) + 1/len(sample2))
    lower = delta - confidence
    upper = delta + confidence
    return delta, lower, upper

def delta_variance_ci(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        level: float = 0.95
        ) -> Tuple[float, float, float]:
    """two sided confidence interval for variance difference of two
    independent variables.

    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second sample.
    level : float in (0, 1), optional
        confidence level between 0 and 1, by default 0.95

    Returns
    -------
    delta : float
        difference of variance of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    
    Notes
    -----
    This function is a ChatGPT solution and therefore does not guarantee
    that this solution is correct.
    """
    var_1 = float(np.var(sample1, ddof=1))
    var_2 = float(np.var(sample2, ddof=1))
    n1 = len(sample1)
    n2 = len(sample2)
    delta = var_1 - var_2
    # Standard error for difference of variances
    se = np.sqrt((2 * var_1**2) / (n1 - 1) + (2 * var_2**2) / (n2 - 1))
    z = norm.ppf(1 - (1 - level) / 2)
    lower = delta - z * se
    upper = delta + z * se
    return delta, lower, upper

def delta_stdev_ci(
        sample1: NumericSample1D,
        sample2: NumericSample1D,
        level: float = 0.95
        ) -> Tuple[float, float, float]:
    """two sided confidence interval for standard deviation difference 
    of two independent variables.

    Parameters
    ----------
    sample1 : NumericSample1D
        A one-dimensional array-like object containing the first sample.
    sample2 : NumericSample1D
        A one-dimensional array-like object containing the second sample.
    level : float in (0, 1), optional
        confidence level between 0 and 1, by default 0.95

    Returns
    -------
    delta : float
        difference of standard deviation of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    
    Notes
    -----
    This function is a ChatGPT solution and therefore does not guarantee
    that this solution is correct.
    """
    s1 = float(np.std(sample1, ddof=1))
    s2 = float(np.std(sample2, ddof=1))
    n1 = len(sample1)
    n2 = len(sample2)
    delta = s1 - s2
    # Delta method for SE of difference of stdevs
    se = np.sqrt((s1**2 / (2 * (n1 - 1))) + (s2**2 / (2 * (n2 - 1))))
    z = norm.ppf(1 - (1 - level) / 2)
    lower = delta - z * se
    upper = delta + z * se
    return delta, lower, upper

def delta_proportions_ci(
        events1: int,
        observations1: int,
        events2: int,
        observations2: int,
        level: float = 0.95
        ) -> Tuple[float, float, float]:
    """Confidence intervals for comparing two independent proportions
    This assumes that we have two independent binomial sample.

    Parameters
    ----------
    events1 : int
        Counted number of events of sample 1.
    observations1 : int
        Total number of observations of sample 1.
    events2 : int
        Counted number of events of sample 2.
    observations2 : int
        Total number of observations of sample 2.
    level : float in (0, 1), optional
        Confidence level, by default 0.95

    Returns
    -------
    delta : float
        Difference of variance of data
    lower : float
        Lower confidence level
    upper : float
        Upper confidence levell
    """
    alpha = confidence_to_alpha(level, two_sided=False)
    delta = events1/observations1 - events2/observations2
    lower, upper = confint_proportions_2indep(
        events1, observations1, events2, observations2, 
        method='wald', compare='diff', alpha=alpha)
    lower = float(lower) # type: ignore
    upper = float(upper) # type: ignore
    return delta, lower, upper

def fit_ci(
        model: RegressionResults,
        level: float = 0.95
        ) -> Tuple[NDArray, NDArray, NDArray]:
    """calculate confidence interval fitted line. Applies to fitted WLS 
    and OLS models, not to general GLS
    
    Parameters
    ----------
    model : statsmodels RegressionResults
        fitted OLS or WLS model
    level : float in (0, 1), optional
        confidence level, by default 0.95
    
    Returns
    -------
    fitted : NDArray
        For coherence with the other functions, the fitted target 
        samples are returned as one-dimensional numpy array,
    lower : NDArray
        Lower confidence limits of fitting line as one-dimensional numpy 
        array.
    upper : NDArray
        Upper confidence limits of fitting line as one-dimensional numpy 
        array.
    
    Notes
    -----
    Using hat_matrix to calculate fit_se only works for fitted values

    This function is based on the summary_table function from the 
    statsmodels.stats.outliers_influence module, see: 
    https://www.statsmodels.org/dev/_modules/statsmodels/stats/outliers_influence.html
    """
    influence = OLSInfluence(model)
    alpha = confidence_to_alpha(level)
    tppf = t.isf(alpha, model.df_resid)
    fit_se = np.sqrt(influence.hat_matrix_diag * model.mse_resid)
    fitted = np.asarray(model.fittedvalues)
    lower = fitted - tppf * fit_se
    upper = fitted + tppf * fit_se
    return fitted, lower, upper

def prediction_ci(
        model: RegressionResults,
        level: float = 0.95
        ) -> Tuple[NDArray, NDArray, NDArray]:
    """calculate confidence interval for prediction and to observe 
    outliers. Applies to fitted WLS and OLS models, not to general GLS.
    
    Parameters
    ----------
    model : statsmodels RegressionResults
        fitted OLS or WLS model
    level : float in (0, 1), optional
        confidence level, by default 0.95
    
    Returns
    -------
    fitted : NDArray
        For coherence with the other functions, the fitted target 
        samples are returned as one-dimensional numpy array,
    lower : NDArray
        Lower confidence limits of prediction as one-dimensional numpy 
        array.
    upper : NDArray
        Upper confidence limits of prediction as one-dimensional numpy 
        array.
    """
    fitted = np.asarray(model.fittedvalues)
    alpha = confidence_to_alpha(level, two_sided=False) # Set two_sided to False here, because the function wls_prediction_std already makes the correction
    lower, upper = wls_prediction_std(model, alpha=alpha)[1:]     # standard error for predicted observation
    lower = np.asarray(lower)
    upper = np.asarray(upper)
    return fitted, lower, upper

def confidence_to_alpha(
        confidence_level: float,
        two_sided: bool = True,
        n_groups: int = 1
        ) -> float:
    """Calculate significance level as alpha risk by given confidence 
    level
    
    Parameters
    ----------
    confidence_level : float in (0, 1)
        level of confidence interval
    two_sided : bool, optional
        True if alpha is to be calculated for a two-sided confidence
        interval, by default True
    n_groups : int
        Used for Bonferroni method. 
        Number of groups to adjust the alpha risk within each group,
        that the total risk is not exceeded, by default 1
    
    Returns
    -------
    alpha : float
        significance level as alpha risk
    """
    assert 0 <= confidence_level <= 1, (
        f'Confidence level {confidence_level} not in (0, 1)')
    assert isinstance(n_groups, int) and n_groups > 0, (
        f'Number of groups must be a unsigned integer > 0')
    sides = 2 if two_sided else 1
    alpha = (1 - confidence_level)/(sides * n_groups)
    return alpha



import sys
import pytest
import matplotlib

import pandas as pd
import statsmodels.api as sm
import matplotlib.pyplot as plt

from typing import Any
from typing import Dict
from typing import List
from pathlib import Path
from matplotlib.text import Text
from pandas.core.frame import DataFrame

sys.path.append(Path(__file__).parent.resolve()) # type: ignore

from daspi import STR
from daspi import COLOR
from daspi import SpecLimits
from daspi import LinearModel
from daspi import load_dataset
from daspi import GageRnRModel
from daspi import GageStudyModel

from daspi.plotlib.chart import Chart
from daspi.plotlib.chart import JointChart
from daspi.plotlib.chart import SingleChart
from daspi.plotlib.chart import MultivariateChart

from daspi.plotlib.plotter import Bar
from daspi.plotlib.plotter import Box
from daspi.plotlib.plotter import Line
from daspi.plotlib.plotter import Stem
from daspi.plotlib.plotter import Pareto
from daspi.plotlib.plotter import Jitter
from daspi.plotlib.plotter import Plotter
from daspi.plotlib.plotter import Scatter
from daspi.plotlib.plotter import Violine
from daspi.plotlib.plotter import Beeswarm
from daspi.plotlib.plotter import Errorbar
from daspi.plotlib.plotter import MeanTest
from daspi.plotlib.plotter import LoessLine
from daspi.plotlib.plotter import StripeLine
from daspi.plotlib.plotter import StripeSpan
from daspi.plotlib.plotter import HideSubplot
from daspi.plotlib.plotter import SkipSubplot
from daspi.plotlib.plotter import SpreadWidth
from daspi.plotlib.plotter import Probability
from daspi.plotlib.plotter import BlandAltman
from daspi.plotlib.plotter import GaussianKDE
from daspi.plotlib.plotter import QuantileBoxes
from daspi.plotlib.plotter import VariationTest
from daspi.plotlib.plotter import ProportionTest
from daspi.plotlib.plotter import CenterLocation
from daspi.plotlib.plotter import TransformPlotter
from daspi.plotlib.plotter import StandardErrorMean
from daspi.plotlib.plotter import ConfidenceInterval
from daspi.plotlib.plotter import ParallelCoordinate
from daspi.plotlib.plotter import GaussianKDEContour
from daspi.plotlib.plotter import LinearRegressionLine
from daspi.plotlib.plotter import CategoricalObservation
from daspi.plotlib.plotter import CapabilityConfidenceInterval

from daspi.plotlib.precast import GageRnRCharts
from daspi.plotlib.precast import GageStudyCharts
from daspi.plotlib.precast import ResidualsCharts
from daspi.plotlib.precast import PairComparisonCharts
from daspi.plotlib.precast import ParameterRelevanceCharts
from daspi.plotlib.precast import BivariateUnivariateCharts
from daspi.plotlib.precast import ProcessCapabilityAnalysisCharts

matplotlib.use("Agg")

savedir = Path(__file__).parent/'charts'
savedir.mkdir(parents=True, exist_ok=True)
df_painkillers = load_dataset('painkillers-dissolution')
"""
Number of observations: 48
Number of features: 8 is composed of 5 factor levels and 3 covariates
Target name: time
    sum of preparation and dissolution
    [time] = seconds
Feature name difinitions:
    employee = 3 levels
        - A
        - B
        - C
    stirrer = 2 levels
        - wooden
        - metal
    brand = 2 levels
        - ZapPain
        - OuchAway
    catalyst = 2 levels
        - False
        - True
    water = 2 levels
        - carbonated
        - tap
    temperature = continuous (covariate)
        [temperature] = °C
    preparation = continuous (covariate)
        [preparation] = s
    dissolution = continuous (covariate)
        [dissolution] = s
"""
df_affairs: DataFrame = sm.datasets.fair.load_pandas().data # type: ignore

df_travel: DataFrame = sm.datasets.modechoice.load_pandas().data # type: ignore
df_travel['mode'] = df_travel['mode'].replace(
    {1: 'air', 2: 'train', 3: 'bus', 4: 'car'})
df_travel['choice'] = df_travel['choice'].replace({0: 'no', 1: 'yes'})
"""
Number of observations: 840 Observations On 4 Modes for 210 Individuals.
Number of variables: 8
Variable name definitions::

    individual = 1 to 210
    mode =
        1 - air
        2 - train
        3 - bus
        4 - car
    choice =
        0 - no
        1 - yes
    ttme = terminal waiting time for plane, train and bus (minutes); 0
           for car.
    invc = in vehicle cost for all stages (dollars).
    invt = travel time (in-vehicle time) for all stages (minutes).
    gc = generalized cost measure:invc+(invt*value of travel time savings)
        (dollars).
    hinc = household income ($1000s).
    psize = traveling group size in mode chosen (number).
    
https://www.statsmodels.org/stable/datasets/generated/modechoice.html
"""

source = Path(__file__).parent/'data'
KW_READ: Dict[str, Any] = dict(sep=';', index_col=0, skiprows=1)
df_dist10: DataFrame = pd.read_csv(
    source/f'dists_10-samples.csv', nrows=10, **KW_READ)
df_dist25: DataFrame = pd.read_csv(
    source/f'dists_25-samples.csv', nrows=25, **KW_READ)

def get_texts(chart: SingleChart | JointChart | MultivariateChart) -> List[Text]:
    return sorted(chart.figure.texts, key=lambda t: t._y, reverse=True) # type: ignore


class TestSingleChart:
    
    fig_title: str = 'SingleChart'
    _sub_title: str = 'Painkillers Dissolution Dataset'
    target_label: str = 'Dissolution time (s)'
    feature_label: str = 'Water temperature (°C)'
    target: str = 'dissolution'
    feature: str = 'temperature'
    kind: str = ''
    info_msg: str = 'Pytest figure, additional info message'
    cat1 = 'employee'
    cat2 ='brand'
    size = 'preparation'

    @property
    def sub_title(self) -> str:
        return f'{self._sub_title}: {self.kind}'
    
    @property
    def file_name(self) -> Path:
        return savedir/f'{self.base}_{self.kind}.png'

    def test_raises(self) -> None:

        with pytest.raises(AssertionError) as err:
            SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
            ).label(info=self.info_msg
            ).plot(Scatter)
        assert 'Cannot call plot() after label()' in str(err.value)

    def test_line_plot(self) -> None:
        self.base = f'{self.fig_title}_line'

        self.kind = 'hue'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1
            ).plot(Line
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True, 
                info=True, 
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 2
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target
        assert texts[3].get_text() == self.feature
        assert STR.TODAY in texts[4].get_text()
        assert STR.USERNAME in texts[4].get_text()
        assert self.info_msg not in info_msg

        self.kind = 'hue_size'
        chart =SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1,
                size=self.size
            ).plot(Line
            ).plot(Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=self.info_msg, 
            ).save(self.file_name
            ).close()
        assert self.file_name.is_file()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 4
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert legend_artists[2].get_children()[0].get_text() == self.size # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.feature_label
        assert STR.TODAY in texts[4].get_text()
        assert STR.USERNAME in texts[4].get_text()
        assert self.info_msg in info_msg
    

    def test_scatter_plot(self) -> None:
        self.base = f'{self.fig_title}_scatter'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature
            ).plot(Scatter
            ).label(
                sub_title=self.sub_title
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 1
        assert chart.label_facets.legend is None
        assert texts[0].get_text() == self.sub_title

        self.kind = 'simple-stripes'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature
            ).plot(Scatter
            ).stripes(
                mean=True,
                median=True,
                control_limits=True,
                spec_limits=SpecLimits(8, 35),
            ).label(
                sub_title=self.sub_title
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 1
        assert chart.label_facets.legend is not None
        assert texts[0].get_text() == self.sub_title

        self.kind = 'simple-stripes-65ci'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature
            ).plot(Scatter
            ).stripes(
                mean=True,
                median=True,
                control_limits=True,
                confidence=0.65,
            ).label(
                sub_title=self.sub_title
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 1
        assert chart.label_facets.legend is not None
        assert texts[0].get_text() == self.sub_title

        self.kind = 'simple-stripes-99ci'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature
            ).plot(Scatter
            ).stripes(
                mean=True,
                median=True,
                control_limits=True,
                confidence=0.99,
            ).label(
                sub_title=self.sub_title
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 1
        assert chart.label_facets.legend is not None
        assert texts[0].get_text() == self.sub_title

        self.kind = 'transposed'
        chart =SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
            ).plot(Scatter, target_on_y=False
            ).label(
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert chart.label_facets.legend is None
        assert texts[0].get_text() == self.sub_title
        assert texts[1].get_text() == self.feature
        assert texts[2].get_text() == self.target


        self.kind = 'hue'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1
            ).plot(Scatter
            ).label(
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4
        assert len(legend_artists) == 2
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert texts[0].get_text() == self.sub_title
        assert texts[1].get_text() == self.target_label
        assert texts[2].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg


        self.kind = 'shape'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                shape=self.cat2
            ).plot(Scatter
            ).label(
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4
        assert len(legend_artists) == 2
        assert legend_artists[0].get_children()[0].get_text() == self.cat2 # type: ignore
        assert texts[0].get_text() == self.sub_title
        assert texts[1].get_text() == self.target_label
        assert texts[2].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg


        self.kind = 'size'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                size=self.size
            ).plot(Scatter
            ).label(
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4
        assert len(legend_artists) == 2
        assert legend_artists[0].get_children()[0].get_text() == self.size # type: ignore
        assert texts[0].get_text() == self.sub_title
        assert texts[1].get_text() == self.target_label
        assert texts[2].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg


        self.kind = 'hue-size'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1,
                size=self.size
            ).plot(Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 4
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert legend_artists[2].get_children()[0].get_text() == self.size # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg

        self.kind = 'hue-shape'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1,
                shape=self.cat2
            ).plot(Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 4
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert legend_artists[2].get_children()[0].get_text() == self.cat2 # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg

        self.kind = 'size-shape'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                size=self.size,
                shape=self.cat2
            ).plot(Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 4
        assert legend_artists[0].get_children()[0].get_text() == self.cat2 # type: ignore
        assert legend_artists[2].get_children()[0].get_text() == self.size # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg

        self.kind = 'full'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1,
                size=self.size,
                shape=self.cat2
            ).plot(Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label, 
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert len(legend_artists) == 6
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert legend_artists[2].get_children()[0].get_text() == self.cat2 # type: ignore
        assert legend_artists[4].get_children()[0].get_text() == self.size # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_pareto_plot(self) -> None:
        self.base = f'{self.fig_title}_pareto'
        with pytest.raises(AssertionError) as err:
            chart = SingleChart(
                    source=df_painkillers,
                    target=self.target,
                    feature=self.cat1,
                    categorical_feature=True
                ).plot(
                    Pareto, method='sum')
        assert 'categorical_feature' in str(err.value)
        
        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
            ).plot(
                Pareto, method='sum'  
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert chart.label_facets.legend is None
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target_label
        assert texts[3].get_text() == self.cat1
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'transposed'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                target_on_y=False
            ).plot(
                Pareto, method='sum'  
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert chart.label_facets.legend is None
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.cat1
        assert texts[3].get_text() == self.target_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
    def test_kde_plot(self) -> None:
        self.base = f'{self.fig_title}_KDE'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                target_on_y=False
            ).plot(
                GaussianKDE,
                hide_axis=None,
                visible_spines=None,
            ).label(
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 2 # feature label should not appear
        assert chart.label_facets.legend is None
        assert texts[0].get_text() == self.sub_title
        assert texts[1].get_text() == self.target

        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                hue=self.cat1,
                target_on_y=True
            ).plot(
                GaussianKDE,
                visible_spines='target',
                hide_axis='feature',
                agreements=(),
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True, 
                info=True
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4 # feature label should not appear
        assert legend_artists[0].get_children()[0].get_text() == self.cat1 # type: ignore
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == self.target
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg not in info_msg
    
    def test_beeswarm_plot(self) -> None:
        self.base = f'{self.fig_title}_beeswarm'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                categorical_feature=True,
            ).plot(
                Beeswarm
            ).plot(
                Violine,
                fill=False
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True,
                target_on_y=False
            ).plot(
                Beeswarm,
            ).plot(
                Violine,
                fill=False
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_quantiles_plot(self) -> None:
        self.base = f'{self.fig_title}_quantiles'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                categorical_feature=True,
            ).plot(
                QuantileBoxes
            ).plot(
                Beeswarm
            ).plot(
                Violine,
                fill=False
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True,
                target_on_y=False
            ).plot(
                QuantileBoxes,
                vary_width=False,
                width=0.1,
                strategy='norm',
            ).plot(
                Beeswarm,
            ).plot(
                Violine,
                fill=False
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

    def test_box_plot(self) -> None:
        self.base = f'{self.fig_title}_box'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                categorical_feature=True,
            ).plot(
                Box,
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'fill'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                categorical_feature=True,
            ).plot(
                Box,
                fill=True,
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True,
                target_on_y=False
            ).plot(
                Box,
                fill=True,
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_jitter_plot(self) -> None:
        self.base = f'{self.fig_title}_jitter'

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                categorical_feature=True,
            ).plot(
                Jitter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True,
                target_on_y=False
            ).plot(
                Jitter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg


    def test_violine_plot(self) -> None:
        self.base = f'{self.fig_title}_violine'

        self.kind = 'mono'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
            ).plot(Violine
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4 # feature label should not appear
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
            ).plot(Violine
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'multiple'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True
            ).plot(Violine, target_on_y=False, agreements=()
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

    def test_errorbar_plots(self) -> None:
        self.base = f'{self.fig_title}_errorbar'

        self.kind = 'sem'
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1, 
                hue=self.cat2,
                dodge=True,
                target_on_y=True
            ).plot(
                StandardErrorMean
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'mean-test'
        n_groups = df_painkillers.groupby(self.cat1).ngroups
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                categorical_feature=True,
                target_on_y=False
            ).plot(
                MeanTest, n_groups=n_groups
            ).label(
                fig_title = self.fig_title,
                sub_title = self.sub_title,
                feature_label = True,
                target_label = self.target_label,    
                info = self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'var-test'
        n_groups = df_painkillers.groupby(self.cat1).ngroups
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                categorical_feature=True,
                target_on_y=False
            ).plot(
                VariationTest,
                kind='variance',
                n_groups=n_groups
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'std-test'
        n_groups = df_painkillers.groupby(self.cat1).ngroups
        chart = SingleChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                categorical_feature=True,
                target_on_y=False,
            ).plot(
                VariationTest,
                kind='stdev',
                show_center=False,
                n_groups=n_groups
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_bland_altman_plot(self) -> None:
        self.base = f'{self.fig_title}_blandaltman'
        
        self.kind = 'simple'
        chart = SingleChart(
                source=load_dataset('shoe-sole'),
                target='wear',
                feature='status',
            ).plot(
                BlandAltman,
                identity='tester',
                feature_axis='mean',
                reverse=True)
        target_label = f'Difference {chart.plots[0].target}'
        sub_title = 'Shoe sole materials dataset'
        feature_label = 'Mean both measurements'
        chart = chart.label(
                fig_title=self.fig_title,
                sub_title=sub_title,
                feature_label=feature_label,
                target_label=target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == sub_title
        assert texts[2].get_text() == target_label
        assert texts[3].get_text() == feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'user'
        chart = SingleChart(
                source=load_dataset('shoe-sole'),
                target='wear',
                feature='status',
            ).plot(
                BlandAltman,
                identity='tester',
                feature_axis='mean',
                agreement=6,
                confidence=0.65,
                lines_same_color=True,
                target_on_y=False,
                reverse=True)
        target_label = f'Difference {chart.plots[0].target}'
        sub_title = 'Shoe sole materials dataset'
        feature_label = 'Mean both measurements'
        chart = chart.label(
                fig_title=self.fig_title,
                sub_title=sub_title,
                feature_label=feature_label,
                target_label=target_label,    
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == sub_title
        assert texts[2].get_text() == feature_label
        assert texts[3].get_text() == target_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_categorical_float_features(self) -> None:
        self.base = f'{self.fig_title}_categorical-float-features'
        
        self.kind = 'simple'
        chart = SingleChart(
                source=df_painkillers,
                target='dissolution',
                feature='temperature',
                categorical_feature=True,
            ).plot(
                Violine,
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 5
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg


class TestJointChart:

    fig_title: str = 'JointChart'
    _sub_title: str = 'Painkillers Dissolution Dataset'
    target_label: str = 'Dissolution time (s)'
    feature_label: str = 'Water temperature (°C)'
    target: str = 'dissolution'
    feature: str = 'temperature'
    kind: str = ''
    info_msg: str = 'Pytest figure, additional info message'
    cat1 = 'employee'
    cat2 ='brand'
    size = 'preparation'

    @property
    def sub_title(self) -> str:
        return f'{self._sub_title}: {self.kind}'
    
    @property
    def file_name(self) -> Path:
        return savedir/f'{self.base}_{self.kind}.png'

    def test_raises(self):
        with pytest.raises(AssertionError) as err:
            JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                nrows=2,
                ncols=2,
                target_on_y=(True, True, False, False)
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=self.target_label,
                info=self.info_msg)
        assert 'Single label not allowed' in str(err.value)
    
        with pytest.raises(AssertionError) as err:
            JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                nrows=2,
                ncols=2,
                target_on_y=(True, True, False, False)
            ).plot(Scatter
            ).label(info=self.info_msg
            ).stripes(mean=True)
        assert 'Cannot call stripes() after label()' in str(err.value)
    

    def test_kde_mean_plots(self) -> None:
        self.base = f'{self.fig_title}_kdes'
        
        self.kind = 'kde-mean'
        n_groups = df_painkillers.groupby([self.cat1, self.cat2]).ngroups
        chart = JointChart(
                source=df_painkillers,
                target=self.target,
                feature=(self.cat1, ''),
                hue=self.cat2,
                ncols=1,
                nrows=2,
                sharex=True,
                dodge=(False, True),
                target_on_y=False,
                height_ratios=[5, 1],
            ).plot(GaussianKDE, show_density_axis=False
            ).plot(MeanTest, n_groups=n_groups
            ).label(
            #     feature_label = (True, True),
            #     target_label = (True, True)
            ).save(self.file_name
            ).close()

    def test_kde_plots(self) -> None:
        self.base = f'{self.fig_title}_kdes'
        
        self.kind = 'mixed-kde'
        chart = JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                hue=self.cat2,
                ncols=1,
                nrows=2,
                sharex=True,
                dodge=(False, True),
                categorical_feature=(False, True),
                target_on_y=False
            ).plot(GaussianKDE, show_density_axis=True
            ).plot(Violine
            ).label(
                feature_label=(False, True),
                target_label=(True, True),
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        legend_artists = chart.label_facets.legend_artists
        yticklabels1 = [t.get_text() for t in chart.axes[1, 0].get_yticklabels()]
        xticklabels0 = [t.get_text() for t in chart.axes[0, 0].get_xticklabels()]
        xticklabels1 = [t.get_text() for t in chart.axes[1, 0].get_xticklabels()]
        assert self.file_name.is_file()
        assert len(texts) == 0 # No text added to figure only to axes
        assert legend_artists[0].get_children()[0].get_text() == self.cat2 # type: ignore
        assert yticklabels1 == sorted(df_painkillers[self.cat1].unique())
        assert bool(xticklabels0) == False
        assert bool(xticklabels1) == True
        assert chart.axes[0, 0].get_xlabel() == ''
        assert chart.axes[1, 0].get_xlabel() == self.target
        assert chart.axes[1, 0].get_ylabel() == self.cat1
    
    def test_probabilities(self) -> None:
        self.base = f'{self.fig_title}_probability'
        
        self.kind = 'norm-prob'
        chart = JointChart(
                df_dist25,
                target='rayleigh',
                feature='',
                nrows=2,
                ncols=2,
                target_on_y=False
            ).plot(Probability, kind='qq'
            ).plot(Probability, kind='pp'
            ).plot(Probability, kind='sq'
            ).plot(Probability, kind='sp'
            ).label(
                fig_title=self.fig_title,
                sub_title='QQ, PP, samples-Q and samples-P',
                target_label=(
                    'norm quantiles', 'norm percentiles',
                    'norm samples', 'norm samples'),
                feature_label=(
                    'theoretical quantiles', 'theoretical percentiles',
                    'theoretical quantiles', 'theoretical percentiles')
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        assert self.file_name.is_file()
        assert len(texts) == 2
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == 'QQ, PP, samples-Q and samples-P'
        assert chart.axes[0, 0].get_ylabel() == 'theoretical quantiles'
        assert chart.axes[0, 1].get_ylabel() == 'theoretical percentiles'
        assert chart.axes[1, 0].get_ylabel() == 'theoretical quantiles'
        assert chart.axes[1, 1].get_ylabel() == 'theoretical percentiles'
        assert chart.axes[0, 0].get_xlabel() == 'norm quantiles'
        assert chart.axes[0, 1].get_xlabel() == 'norm percentiles'
        assert chart.axes[1, 0].get_xlabel() == 'norm samples'
        assert chart.axes[1, 1].get_xlabel() == 'norm samples'
        for l in [t.get_text() for t in chart.axes[0, 0].get_yticklabels()]:
            assert '%' not in l
        for l in [t.get_text() for t in chart.axes[0, 1].get_yticklabels()]:
            assert '%' in l
        for l in [t.get_text() for t in chart.axes[1, 0].get_yticklabels()]:
            assert '%' not in l
        for l in [t.get_text() for t in chart.axes[1, 1].get_yticklabels()]:
            assert '%' in l
        for l in [t.get_text() for t in chart.axes[0, 0].get_xticklabels()]:
            assert '%' not in l
        for l in [t.get_text() for t in chart.axes[0, 1].get_xticklabels()]:
            assert '%' in l
        for l in [t.get_text() for t in chart.axes[1, 0].get_xticklabels()]:
            assert '%' not in l
        for l in [t.get_text() for t in chart.axes[1, 1].get_xticklabels()]:
            assert '%' not in l

        self.kind = 'dists-prob'
        target = tuple(df_dist25.columns.to_list()[1:])
        target_labels = tuple(f'{d} quantiles' for d in target)
        feature_label = 'theoretical quantiles'
        chart = JointChart(
                df_dist25,
                target=target,
                feature='',
                nrows=3,
                ncols=3
            )
        for distribution in target:
            chart.plot(Probability, dist=distribution, kind='qq')
        chart.label(
                fig_title=self.fig_title,
                sub_title='QQ for different distributions',
                target_label=target_labels,
                feature_label=feature_label,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4
        for ax, ylabel in zip(chart.axes.flat, target_labels):
            assert ax.get_xlabel() == ''
            assert ax.get_ylabel() == ylabel
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == 'QQ for different distributions'
        assert texts[2].get_text() == feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

    def test_regression_joint(self) -> None:
        self.base = f'{self.fig_title}_regression'

        self.kind = 'kde'
        target_labels = ('', '', 'In vehicle time (s)', '')
        feature_labels = ('', '', 'In vehicle cost ($)', '')
        chart = JointChart(
                source=df_travel,
                target=('invc', '', 'invt', 'invt'),
                feature=('', '', 'invc', ''),
                target_on_y=(False, False, True, True),
                hue='mode',
                nrows=2,
                ncols=2,
                sharex='col',
                sharey='row',
                width_ratios=[5, 1],
                height_ratios=[1, 5],
                stretch_figsize=False
        ).plot(GaussianKDE, show_density_axis=False
        ).plot(HideSubplot
        ).plot(LinearRegressionLine, show_scatter=True, show_fit_ci=True
        ).plot(GaussianKDE, show_density_axis=False
        ).label(
            feature_label=feature_labels,
            target_label=target_labels,
            fig_title=self.fig_title,
            sub_title='Traveling Mode Dataset',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == 'Traveling Mode Dataset'
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        for ax, flabel, tlabel in zip(chart.axes.flat, feature_labels, target_labels):
            assert ax.get_xlabel() == flabel
            assert ax.get_ylabel() == tlabel

    def test_bar_plots(self) -> None:
        self.base = f'{self.fig_title}_bar'

        self.kind = 'stacking'
        target_labels = (
            'Counted observed dissolution time',
            'Dissolution time (s)')
        feature_label = 'Employee'
        chart = JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat2,
                nrows=2,
                ncols=1,
                hue=self.cat1,
                categorical_feature=True,
                target_on_y=(False, False),
                dodge=(False, True)
            ).plot(Bar, method='count'
            ).plot(Bar, method='sum'
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=feature_label, 
                target_label=target_labels,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 4
        for ax, xlabel in zip(chart.axes.flat, target_labels):
            assert ax.get_xlabel() == xlabel
            assert ax.get_ylabel() == ''
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert texts[2].get_text() == feature_label
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

    def test_pareto_plot(self) -> None:
        self.base = f'{self.fig_title}_pareto'
        with pytest.raises(AssertionError) as err:
            JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                nrows=3,
                ncols=2,
                sharex='col',
            ).plot(Pareto, method='sum')
        assert 'shared with other axes' in str(err.value)

        self.kind = 'marked'
        chart = JointChart(
                source=df_painkillers,
                target=self.target,
                feature=self.cat1,
                nrows=2,
                ncols=2,
                target_on_y=(True, True, False, False)
            ).plot(Pareto, method='sum', highlight='C', highlighted_as_last=False
            ).plot(Pareto, method='sum', highlight='B', highlight_color=COLOR.GOOD
            ).plot(Pareto, method='sum', highlight='C', highlighted_as_last=False
            ).plot(Pareto, method='sum', highlight='B', highlight_color=COLOR.GOOD
            ).stripes(mean=True
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=(True, True, True, True),
                target_label=tuple([self.target_label]*4),
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        assert self.file_name.is_file()
        assert len(texts) == 3 # Feature and target labels are not as texts in figure here
        for i, ax in enumerate(chart.axes.flat):
            if i < 2:
                assert ax.get_xlabel() == self.cat1
                assert ax.get_ylabel() == self.target_label
            else:
                assert ax.get_xlabel() == self.target_label
                assert ax.get_ylabel() == self.cat1
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg


class TestMultivariateChart:

    fig_title: str = 'MultivariateChart'
    _sub_title: str = 'Traveling Mode Dataset'

    @property
    def sub_title(self) -> str:
        return f'{self._sub_title}: {self.kind}'
    
    @property
    def file_name(self) -> Path:
        return savedir/f'{self.base}_{self.kind}.png'

    def test_full(self) -> None:
        self.base = f'{self.fig_title}'

        self.kind = 'full'
        chart = MultivariateChart(
                source=df_affairs,
                target='affairs',
                feature='yrs_married', 
                hue='rate_marriage',
                size='age',
                shape='educ',
                col='religious',
                row='children',
                stretch_figsize=(2.0, 2.5)
            ).plot(Scatter
            ).label(
                fig_title='Multiple Variate Chart',
                sub_title='Affairs R Dataset',
                feature_label='Years of marriage',
                target_label='Amount of affairs',
                row_title='Amount of children',
                col_title='How religious',
                info='pytest figure')
        chart.axes[0].set(ylim=(-5, 60))
        chart.save(self.file_name).close()
        assert self.file_name.is_file()

        self.base = f'{self.fig_title}'
        self.kind = 'full-stripes'
        chart = MultivariateChart(
                source=df_affairs,
                target='affairs',
                feature='yrs_married', 
                hue='rate_marriage',
                size='age',
                shape='educ',
                col='religious',
                row='children',
                stretch_figsize=(2.0, 2.5)
            ).plot(Scatter
            ).stripes(
                mean=True,
                control_limits=True,
                confidence=0.95,
            ).label(
                fig_title='Multiple Variate Chart',
                sub_title='Affairs R Dataset',
                feature_label='Years of marriage',
                target_label='Amount of affairs',
                row_title='Amount of children',
                col_title='How religious',
                info='pytest figure')
        chart.axes[0].set(ylim=(-5, 60))
        chart.save(self.file_name).close()
        assert self.file_name.is_file()


class TestPrecasts:

    fig_title: str = 'Precasts'
    _sub_title: str = 'Painkillers Dissolution Dataset'
    target_label: str = 'Dissolution time (s)'
    feature_label: str = 'Water temperature (°C)'
    target: str = 'dissolution'
    feature: str = 'temperature'
    kind: str = ''
    base: str = ''
    info_msg: str = 'Pytest figure, additional info message'
    cat1 = 'employee'
    cat2 ='brand'
    size = 'preparation'

    @property
    def sub_title(self) -> str:
        return f'{self._sub_title}: {self.kind}'
    
    @property
    def file_name(self) -> Path:
        return savedir/f'{self.base}_{self.kind}.png'
    
    @pytest.fixture
    def df_painkillers(self) -> pd.DataFrame:
        return load_dataset('painkillers-dissolution')
    
    def test_bivariate_univariate_charts(self, df_painkillers: pd.DataFrame) -> None:
        
        self.base = f'{self.fig_title}_bivariate-univariate-charts'
        self.kind = 'simple'
        chart = BivariateUnivariateCharts(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat1
            ).plot_univariates(
                GaussianKDE
            ).plot_bivariate(
                LinearRegressionLine
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert xlabels == ('', '', self.feature, '')
        assert ylabels == ('', '', self.target, '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.base = f'{self.fig_title}_bivariate-univariate-charts'
        self.kind = 'smoothed'
        chart = BivariateUnivariateCharts(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat2
            ).plot_univariates(
                GaussianKDE
            ).plot_bivariate(
                LoessLine, show_ci=True
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=True,
                target_label=True,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert xlabels == ('', '', self.feature, '')
        assert ylabels == ('', '', self.target, '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
        
        self.kind = 'mean-test'
        _title = '95 % confidence interval of mean'
        n_groups = df_painkillers.groupby(self.cat2).ngroups
        chart = BivariateUnivariateCharts(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat2,
                dodge_univariates=True,
                stretch_figsize=False,
            ).plot_univariates(
                MeanTest, n_groups=n_groups
            ).plot_bivariate(
                LinearRegressionLine, show_fit_ci=True
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label,
                axes_titles=(_title, '', '', ''),
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert xlabels == ('', '', self.feature_label, '')
        assert ylabels == ('', '', self.target_label, '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'contour-mean-test'
        _title = '95 % confidence interval of mean'
        n_groups = df_painkillers.groupby(self.cat2).ngroups
        chart = BivariateUnivariateCharts(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat2,
                dodge_univariates=True,
                stretch_figsize=False,
            ).plot_univariates(
                MeanTest, n_groups=n_groups
            ).plot_bivariate(
                GaussianKDEContour
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label,
                axes_titles=(_title, '', '', ''),
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert xlabels == ('', '', self.feature_label, '')
        assert ylabels == ('', '', self.target_label, '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        self.kind = 'user'
        _title = '95 % confidence interval of mean'
        n_groups = df_painkillers.groupby(self.cat2).ngroups
        chart = BivariateUnivariateCharts(
                source=df_painkillers,
                target=self.target,
                feature=self.feature,
                hue=self.cat2,
                dodge_univariates=True,
                stretch_figsize=False,
            ).plot_univariates(
                MeanTest, n_groups=n_groups
            ).plot_bivariate(
                GaussianKDEContour, fill=False, fade_outers=False
            ).plot_univariates(
                QuantileBoxes, strategy='fit'
            ).plot_bivariate(
                Scatter
            ).label(
                fig_title=self.fig_title,
                sub_title=self.sub_title,
                feature_label=self.feature_label,
                target_label=self.target_label,
                axes_titles=(_title, '', '', ''),
                info=self.info_msg
            ).save(self.file_name
            ).close()
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == self.fig_title
        assert texts[1].get_text() == self.sub_title
        assert xlabels == ('', '', self.feature_label, '')
        assert ylabels == ('', '', self.target_label, '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg

        fitted_dists = []
        for plot in chart.plots:
            if isinstance(plot, QuantileBoxes):
                fitted_dists.append(plot.estimation.dist.name)
        fitted_dists.sort()
        assert fitted_dists == ['foldnorm', 'gamma', 'logistic', 'norm']

    def test_process_capabity_analysis_charts(self) -> None:
        
        df = load_dataset('drop_card')
        spec_limits = SpecLimits(0, float(df.loc[0, 'usl'])) # type: ignore
        target = 'distance'
        fig_title='Process Capability Analysis'
        sub_title='Drop Card Experiment'
        target_label='Distance (cm)'

        self.base = f'{self.fig_title}_process-capabity-analysis-charts'
        self.kind = 'hue'
        chart = ProcessCapabilityAnalysisCharts(
                source=df,
                target=target,
                spec_limits=spec_limits,
                hue='method'
            ).plot(
            ).stripes(
            ).label(
                fig_title=fig_title,
                sub_title=sub_title,
                target_label=target_label,
                info=self.info_msg
            ).save(self.file_name
            ).close()
        
        texts = get_texts(chart)
        info_msg = texts[-1].get_text()
        xlabels = tuple(ax.get_xlabel() for ax in chart.axes.flat)
        ylabels = tuple(ax.get_ylabel() for ax in chart.axes.flat)
        assert self.file_name.is_file()
        assert len(texts) == 3
        assert texts[0].get_text() == fig_title
        assert texts[1].get_text() == sub_title
        assert xlabels == (
            STR['charts_flabel_observed'], target_label, target_label, 
            STR['cpk'], STR['cp'])
        assert ylabels == (
            target_label, STR['charts_flabel_quantiles'], '', '', '')
        assert STR.TODAY in info_msg
        assert STR.USERNAME in info_msg
        assert self.info_msg in info_msg
    
    def test_parameter_relevance_charts(self) -> None:
        df = load_dataset('painkillers-dissolution')
        model = LinearModel(
            source=df,
            target='dissolution',
            features=['employee', 'stirrer', 'brand', 'catalyst', 'water'],
            covariates=['temperature', 'preparation'],
            order=2)
        df_gof = pd.concat(model.recursive_elimination())
        
        self.base = f'{self.fig_title}_parameter_relevance_charts'
        self.kind = 'basic'
        chart = ParameterRelevanceCharts(model
            ).plot(
            ).stripes(
            ).label(info=True
            ).save(self.file_name)
        assert self.file_name.is_file()
    
    def test_pair_comparison_charts(self) -> None:
        df = load_dataset('shoe-sole')
        
        self.base = f'{self.fig_title}_pair_comparison_charts'
        self.kind = 'basic'
        chart = PairComparisonCharts(
                source=df,
                target='wear',
                feature='status',
                identity='tester'
            ).plot(
            ).label(
                info=True
            ).save(self.file_name)
        assert self.file_name.is_file()

    def test_residual_charts(self) -> None:
        df = load_dataset('painkillers-dissolution')
        model = LinearModel(
            source=df,
            target='dissolution',
            features=['employee', 'stirrer', 'brand', 'catalyst', 'water'],
            covariates=['temperature', 'preparation'],
            order=2)
        df_gof = pd.concat(model.recursive_elimination())
        
        self.base = f'{self.fig_title}_residual_charts'
        self.kind = 'basic'
        chart = ResidualsCharts(model
            ).plot(
            ).stripes(
            ).label(info=True
            ).save(self.file_name)
        assert self.file_name.is_file()

    def test_gage_study_model(self) -> None:
        df = load_dataset('grnr_layer_thickness')
        gage = GageStudyModel(
            source=df,
            target='result_gage',
            reference='reference',
            u_cal=df['U_cal'][0],
            tolerance=df['tolerance'][0],
            resolution=df['resolution'][0],
            bias_corrected=True,)
        
        self.base = f'{self.fig_title}_gage_study_model'
        self.kind = 'basic'
        chart = GageStudyCharts(gage, stretch_figsize=1.5
            ).plot(
            ).stripes(
            ).label(
            ).save(self.file_name)
        assert self.file_name.is_file()
    
    def test_gage_rnr_charts(self) -> None:
        df = load_dataset('grnr_layer_thickness')
        gage = GageStudyModel(
            source=df,
            target='result_gage',
            reference='reference',
            u_cal=df['U_cal'][0],
            tolerance=df['tolerance'][0],
            resolution=df['resolution'][0])
        model = GageRnRModel(
            source=df,
            target='result_rnr',
            part='part',
            gage=gage,
            u_av='operator')
        
        self.base = f'{self.fig_title}_gage_rnr_model'
        self.kind = 'basic'
        chart = GageRnRCharts(model, stretch_figsize=True
            ).plot(
            ).stripes(
            ).label(
            ).save(self.file_name)
        assert self.file_name.is_file()


class TestFormattingFeatures:
    """Test new formatting features: formatters, angles, and alignments."""
    
    fig_title: str = 'Formatting Features Test'
    info_msg: str = 'Pytest figure for formatting features'
    kind: str = ''
    base: str = ''

    @property
    def file_name(self) -> Path:
        return savedir/f'{self.base}_{self.kind}.png'

    def test_formatters_basic(self) -> None:
        """Test basic formatter functionality."""
        self.base = f'{self.fig_title.replace(" ", "_")}_formatters'
        self.kind = 'basic'
        
        # Custom formatter function
        def custom_formatter(x, pos):
            return f'${x:.1f}k'
        
        chart = SingleChart(
            source=load_dataset('mpg'),
            target='cty',
            feature='displ'
        ).plot(Scatter
        ).label(
            fig_title=self.fig_title,
            sub_title='Basic Formatter Test',
            feature_formatter=custom_formatter,
            target_formatter=lambda x, pos: f'{x:.0f} MPG',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_formatters_scientific(self) -> None:
        """Test scientific notation formatter."""
        self.base = f'{self.fig_title.replace(" ", "_")}_formatters'
        self.kind = 'scientific'
        
        chart = SingleChart(
            source=load_dataset('mpg'),
            target='cty',
            feature='hwy'
        ).plot(Scatter
        ).label(
            fig_title=self.fig_title,
            sub_title='Scientific Notation Test',
            feature_formatter='{:.2e}',
            target_formatter='{:.1f}',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_angles_basic(self) -> None:
        """Test basic angle rotation."""
        self.base = f'{self.fig_title.replace(" ", "_")}_angles'
        self.kind = 'basic'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day'
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='Basic Angle Test',
            feature_angle=45,
            target_angle=90,
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_alignments_basic(self) -> None:
        """Test basic alignment options."""
        self.base = f'{self.fig_title.replace(" ", "_")}_alignments'
        self.kind = 'basic'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day'
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='Basic Alignment Test',
            feature_align='right',
            target_align='center',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_combined_features(self) -> None:
        """Test combining formatters, angles, and alignments."""
        self.base = f'{self.fig_title.replace(" ", "_")}_combined'
        self.kind = 'all_features'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day'
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='Combined Features Test',
            feature_formatter=lambda x, pos: str(x).upper(),
            target_formatter='${:.2f}',
            feature_angle=30,
            target_angle=45,
            feature_align='left',
            target_align='right',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_multivariate_formatting(self) -> None:
        """Test formatting features with MultivariateChart."""
        self.base = f'{self.fig_title.replace(" ", "_")}_multivariate'
        self.kind = 'formatting'
        
        chart = MultivariateChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='size',
            hue='sex',
            col='time'
        ).plot(Scatter
        ).label(
            fig_title=self.fig_title,
            sub_title='Multivariate Formatting Test',
            feature_formatter='Size: {:.0f}',
            target_formatter='${:.2f}',
            feature_angle=15,
            target_angle=30,
            feature_align='center',
            target_align='left',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_orientation_mapping(self) -> None:
        """Test formatting with different orientations."""
        self.base = f'{self.fig_title.replace(" ", "_")}_orientation'
        self.kind = 'mapping'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day',
            target_on_y=False
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='Orientation Mapping Test',
            feature_formatter=lambda x, pos: f'Day: {x}',
            target_formatter='${:.1f}',
            feature_angle=60,
            target_angle=0,
            feature_align='right',
            target_align='center',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_percentage_formatter(self) -> None:
        """Test percentage formatting."""
        self.base = f'{self.fig_title.replace(" ", "_")}_percentage'
        self.kind = 'formatter'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='tip',
            feature='total_bill'
        ).plot(Scatter
        ).label(
            fig_title=self.fig_title,
            sub_title='Percentage Formatter Test',
            target_formatter='{:.1%}',
            feature_formatter='${:.0f}',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_callable_formatters(self) -> None:
        """Test callable formatters with different signatures."""
        self.base = f'{self.fig_title.replace(" ", "_")}_callable'
        self.kind = 'formatters'
        
        def business_formatter(value, position=None):
            """Business-style number formatter."""
            if value >= 1000000:
                return f'{value/1000000:.1f}M'
            elif value >= 1000:
                return f'{value/1000:.1f}K'
            else:
                return f'{value:.0f}'
        
        chart = SingleChart(
            source=load_dataset('mpg'),
            target='cty',
            feature='displ'
        ).plot(Scatter
        ).label(
            fig_title=self.fig_title,
            sub_title='Callable Formatters Test',
            feature_formatter=business_formatter,
            target_formatter=lambda x, pos: f'{x:.1f} MPG',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_extreme_angles(self) -> None:
        """Test extreme angle values."""
        self.base = f'{self.fig_title.replace(" ", "_")}_extreme'
        self.kind = 'angles'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day'
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='Extreme Angles Test',
            feature_angle=90,
            target_angle=-45,
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()

    def test_all_alignments(self) -> None:
        """Test all alignment options."""
        self.base = f'{self.fig_title.replace(" ", "_")}_all'
        self.kind = 'alignments'
        
        chart = SingleChart(
            source=load_dataset('tips'),
            target='total_bill',
            feature='day'
        ).plot(Violine
        ).label(
            fig_title=self.fig_title,
            sub_title='All Alignments Test',
            feature_align='left',
            target_align='right',
            info=self.info_msg
        ).save(self.file_name
        ).close()
        assert self.file_name.is_file()



from importlib.metadata import version
from importlib.metadata import PackageNotFoundError

try:
    __version__ = version('daspi')
except PackageNotFoundError:
    __version__ = 'local development version'



import random

import numpy as np

from math import pi
from math import sin
from scipy import stats
from typing import Any
from typing import Self
from typing import Tuple
from typing import Literal
from dataclasses import dataclass
from numpy.typing import NDArray
from pandas.core.series import Series

from .._typing import FloatOrArray


def calculate_agreement_and_k(agreement: int | float) -> Tuple[float, float]:
    """Calculate agreement and coverage factor k from input agreement.
    
    This utility function standardizes the calculation of agreement and 
    coverage factor k from various input formats. It handles both 
    standard deviation multipliers (integers/floats >= 1) and percentile 
    values (floats between 0 and 1).
    
    Parameters
    ----------
    agreement : int | float
        The agreement value, either as:
        - A percentage (0.0 < agreement <= 1.0) indicating the
          acceptable proportion for the spread.
        - A multiple of the standard deviation (agreement >= 1).
    
    Returns
    -------
    Tuple[float, float]
        A tuple containing (agreement, k) where:
        - agreement: The standardized agreement value (always >= 1)
        - k: The coverage factor
    
    Raises
    ------
    AssertionError
        If agreement <= 0 or not in valid range.
    
    Examples
    --------
    >>> calculate_agreement_and_k(6)  # 6σ
    (6.0, 3.0)
    >>> calculate_agreement_and_k(0.95)  # 95% percentile
    (3.919927969080108, 1.959963984540054)
    """
    assert agreement > 0, (
        'Agreement must be set as a percentage (0.0 < agreement <= 1.0) '
        + 'or as a multiple of the standard deviation (agreement >= 1), '
        + f'got {agreement}.')
    
    is_percentile = (
        agreement < 1 or (agreement == 1 and isinstance(agreement, float)))
    if is_percentile:
        k = float(stats.norm.ppf((1 + agreement) / 2))
        agreement = 2 * k
    else:
        k = agreement / 2
    
    return float(agreement), float(k)


__all__ = [
    'SpecLimits',
    'Specification',
    'RandomProcessValue',
    'Binning',
    'round_to_nearest',
    'inclination_displacement',
    'calculate_agreement_and_k']


@dataclass(frozen=True)
class SpecLimits:
    """Class to hold the limits of a parameter specification.

    Parameters
    ----------
    lower : float, optional
        The lower limit of the specification. Default is -inf.
    upper : float, optional
        The upper limit of the specification. Default is inf.
    """
    lower: float = float('-inf')
    upper: float = float('inf')

    def __post_init__(self) -> None:
        assert  self.lower <= self.upper, (
            'Lower limit must be less than or equal to the upper limit.')

    @property
    def is_unbounded(self) -> bool:
        """Check if any of lower or upper is -inf or inf."""
        return self.lower == float('-inf') or self.upper == float('inf')

    @property
    def are_both_finite(self) -> bool:
        """Check if both lower and upper are finite values."""
        return (self.lower > float('-inf')) and (self.upper < float('inf'))

    @property
    def both_unbounded(self) -> bool:
        """Check if both lower and upper are -inf and inf, respectively."""
        return self.lower == float('-inf') and self.upper == float('inf')
    
    @property
    def tolerance(self) -> float:
        """Tolerance range, returns the difference between upper and 
        lower limits. (read-only)"""
        return self.upper - self.lower
    
    @property
    def nominal(self) -> float:
        """Nominal value, returns the average of upper and lower limits
        (read-only)."""
        return (self.upper + self.lower) / 2
    
    def to_tuple(self) -> Tuple[float, float]:
        """Returns the lower and upper limits as a tuple.

        Returns
        -------
        Tuple[float, float]
            A tuple containing the lower and upper limits in order.
        """
        return (self.lower, self.upper)

    def __contains__(self, value: float) -> bool:
        """Check if a given value is within the specified limits.

        Parameters
        ----------
        value : float
            The value to check against the limits.

        Returns
        -------
        bool
            True if the value is within the limits, False otherwise.
        """
        return self.lower <= value <= self.upper
    
    def __eq__(self, other: Any) -> bool:
        """
        Checks if this SpecLimits instance is equal to another 
        SpecLimits instance or a tuple.

        Parameters
        ----------
        other : SpecLimits, list or tuple
            The object to compare with this instance.

        Returns
        -------
        bool
            True if equal, False otherwise.
        """
        if isinstance(other, SpecLimits):
            return self.to_tuple() == other.to_tuple()
        elif isinstance(other, (tuple, list)) and len(other) == 2:
            return self.to_tuple() == tuple(other)
        return False
    

class Specification:
    """Class for defining the boundaries of a parameter specification. 
    
    This class is used to create constants. For this reason, all 
    properties are uppercase.

    Parameters
    ----------
    limits : SpecLimits | Tuple[float, float] | None, optional
        The limits of the specification. The limits are automatically 
        calculated if only tolerance is given. Default is None.
    tolerance : float or None, optional
        The tolerance range of the specification. Default is None.
    nominal : float or None, optional
        The nominal value of the specification. Default is None.
    n_digits : int, optional
        The number of decimal places to counteract the rounding error 
        of floating-point numbers. Default is 10.

    Notes
    -----
    The nominal and tolerance are automatically calculated if only 
    limits are provided. The limits and nominal are automatically 
    calculated if only tolerance is provided.
    - Tolerance is calculated as the difference between the upper and 
      lower limits.
    - Nominal is calculated as the average of the lower and upper limits.
    - Limits are calculated as the nominal +/- tolerance / 2.

    Raises
    ------
    AssertionError
        If the lower limit is greater than the upper limit.
    AssertionError
        If limits, tolerance and nominal are not provided.

    Examples
    --------
    Create a specification with limits:
    ```python
    Specification(limits=(0, 1))
    ```

    ```
    Specification(limits=(0, 1), tolerance=1.0, nominal=0.5)
    ```
    
    Create a specification with tolerance and nominal:
    ```python
    Specification(tolerance=0.1, nominal=0.5)
    ```
    
    ```
    Specification(limits=(0.45, 0.55), tolerance=0.1, nominal=0.5)
    ```
    """

    __slots__ = ('_limits', '_tolerance', '_nominal', '_resolution')

    _limits: SpecLimits
    """The limits of the specification parameter."""

    _tolerance: float
    """The tolerance range of the specification."""

    _nominal: float
    """The nominal value of the specification."""

    _resolution: int
    """The resolution as 1e-n_digits."""

    def __init__(
            self,
            *,
            limits: SpecLimits | Tuple[float, float] | None = None, 
            tolerance: float | None = None,
            nominal: float | None = None,
            n_digits: int = 10
            ) -> None:
        assert n_digits > 0, 'n_digits must be greater than 0.'

        if isinstance(limits, tuple):
            limits = SpecLimits(*limits)
        elif limits is None:
            assert tolerance is not None and nominal is not None, (
                'Either limits or tolerance and nominal must be provided.')
            limits = SpecLimits(
                round(nominal - tolerance / 2, n_digits),
                round(nominal + tolerance / 2, n_digits))

        if tolerance is None:
            assert limits.are_both_finite, (
                'If tolerance is not provided, limits must be finite.')
            tolerance = round(limits.upper - limits.lower, n_digits)
        
        if nominal is None:
            assert limits.are_both_finite, (
                'If nominal is not provided, limits must be finite.')
            nominal = round(limits.lower + tolerance / 2, n_digits)

        self._limits = limits
        self._tolerance = tolerance
        self._nominal = nominal
        self._resolution = 1 ** -n_digits
    
    @property
    def LIMITS(self) -> SpecLimits:
        """The lower and upper limits of the specification (read-only)."""
        return self._limits
    
    @property
    def TOLERANCE(self) -> float:
        """The tolerance range of the specification."""
        return self._tolerance
    
    @property
    def NOMINAL(self) -> float:
        """The nominal value of the specification."""
        return self._nominal
    
    @property
    def RESOLUTION(self) -> float:
        """The resolution of the specification."""
        return self._resolution
    
    @property
    def LOWER(self) -> float:
        """The lower limit of the specification."""
        return self.LIMITS.lower
    
    @property
    def UPPER(self) -> float:
        """The upper limit of the specification."""
        return self.LIMITS.upper
    
    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(limits={self.LIMITS.to_tuple()}, '
            f'tolerance={self.TOLERANCE}, nominal={self.NOMINAL})')

    def __str__(self) -> str:
        return (
            f'{self.__class__.__name__}('
            f'limits={self.LIMITS.to_tuple()}, tolerance={self.TOLERANCE}, '
            f'nominal={self.NOMINAL})')


class RandomProcessValue:
    """Class to generate a random process value.

    This class generates random values based on the specified
    distribution and parameters. The generated values are within the 
    specified limits and tolerance. The class supports five types 
    of distributions: normal, uniform, circular, coaxial, and perpendicular. 
    The generated values can be clipped to the range defined by the tolerance. 
    The class also provides a method to generate an array of random values.
    
    Parameters
    ----------
    specification : Specification
        The specification for which the random value will be generated.
    dist : {'normal', 'uniform', 'circular', 'coaxial', 'perpendicular'}
        The distribution from which the random value will be generated.
    clip : bool, optional
        Whether to clip the generated value to the range defined by the
        tolerance. Default is False.
    agreement : int | float, optional
        Specify the tolerated process variation for which the 
        scale factor is calculated. 
        - If int, the spread is determined using the normal 
          distribution agreement*σ, 
          e.g. agreement = 6 -> 6*σ ~ covers 99.75% of the data. 
        - If float, the value must be between 0 and 1. This value is
          then interpreted as the acceptable proportion for the 
          spread, e.g. 0.9973 (which corresponds to ~ 6σ)
        Default is 6.
    
    Examples
    --------
    Generate a random value from a uniform distribution:
    
    ```python
    import daspi as dsp

    PARAM_I = dsp.Specification(limits=(5, 10))
    PARAM_II = dsp.Specification(tolerance=0.1, nominal=3)

    rpv = RandomProcessValue(PARAM_I, 'uniform')
    value = rpv()
    ```

    Generate an array from a normal distribution with clipping:

    ```python
    rpv = RandomProcessValue(PARAM_II, 'normal', clip=True)
    array = rpv.generate(100_000)
    ```
    
    Generate values with custom agreement (4σ instead of default 6σ):
    
    ```python
    rpv = RandomProcessValue(PARAM_II, 'normal', agreement=4)
    value = rpv()
    print(f'Scale factor: {rpv.scale}')  # TOLERANCE / 4
    ```
    
    Use percentile-based agreement (95% coverage):
    
    ```python
    rpv = RandomProcessValue(PARAM_II, 'normal', agreement=0.95)
    value = rpv()
    print(f'Coverage factor k: {rpv.k}')  # ~1.96 for 95% confidence
    ```
    """

    _allowed_dists: Tuple[str, ...] = (
        'normal', 'uniform', 'circular', 'coaxial', 'perpendicular')
    """Allowed distributions for the RandomProcessValue class."""

    def __init__(
            self,
            specification: Specification,
            dist: Literal['normal', 'uniform', 'circular', 'coaxial', 'perpendicular'],
            clip: bool = False,
            agreement: int | float = 6,
            ) -> None:
        assert dist in self._allowed_dists, (
            f'Distribution must be one of {self._allowed_dists}, got {dist}.')
        
        self.specification = specification
        self.dist = dist
        self.clip_within_tolerance = clip
        self._agreement: float
        self._k: float
        self.agreement = agreement
    
    @property
    def agreement(self) -> float:
        """Get the agreement multiplier for the σ (standard deviation).
        
        The agreement is defined as twice the coverage factor `k`. 
        When setting the agreement using a percentile, provide the 
        acceptable proportion for the spread, such as 0.9973, 
        which corresponds to approximately 6σ (six standard deviations).
        The agreement value must be specified as either:
        - A percentage (0.0 < agreement <= 1.0) indicating the
          acceptable proportion for the spread.
        - A multiple of the standard deviation (agreement >= 1).
        
        Default is 6 (corresponding to 6σ or k=3).
        """
        return self._agreement
    
    @agreement.setter
    def agreement(self, agreement: int | float) -> None:
        self._agreement, self._k = calculate_agreement_and_k(agreement)
    
    @property
    def k(self) -> float:
        """Get the coverage factor `k` used in scale calculations (read-only).
        
        This property returns the coverage factor, which is a multiplier 
        used to determine the scale based on the agreement. The value of `k` 
        is typically set to reflect the desired confidence level.
        """
        return self._k
    
    @property
    def scale(self) -> float:
        """Scale factor for the distribution based on agreement."""
        return self.specification.TOLERANCE / self.agreement
    
    @property
    def loc(self) -> float:
        """Location parameter of the distribution."""
        return self.specification.NOMINAL
    
    @property
    def lower(self) -> float:
        """Lower bound of the distribution."""
        return self.specification.LOWER
    
    @property
    def upper(self) -> float:
        """Upper bound of the distribution."""
        return self.specification.UPPER
    
    @staticmethod
    def clip(value: float, min_value: float, max_value: float) -> float:
        """Clip a value between a minimum and maximum value.
        
        This function ensures that the provided value is within the 
        specified range, clipping it if it falls outside the range.
        
        Parameters
        ----------
        value : float
            The value to be clipped.
        min_value : float
            The minimum value of the range.
        max_value : float
            The maximum value of the range.
        
        Returns
        -------
        float
            The clipped value.
        """
        return max(min(value, max_value), min_value)

    def normal(self) -> float:
        """Generate a random value within the specified tolerance range.
        
        This function generates a random value within a range defined by
        the provided nominal and tolerance. The generated value is
        calculated by a normal distribution with a mean of the nominal
        and a standard deviation equal to the tolerance divided by 6.
        
        Returns
        -------
        float
            The generated random value within the specified tolerance 
            range.
        """
        value = random.normalvariate(self.loc, self.scale)
        if self.clip_within_tolerance:
            value = self.clip(value, self.lower, self.upper)
        return value
    
    def uniform(self) -> float:
        """Generate a random value within the specified tolerance range.
        
        This method generates a random value within the specified 
        tolerance range using a uniform distribution.
        
        Returns
        -------
        float
            The generated random value within the specified tolerance 
            range.
        """
        value = random.uniform(self.lower, self.upper)
        return value
    
    def circular(self) -> float:
        """Generate a randomized offset based on a circular distribution.

        This function computes a randomized value by applying a circular 
        offset to the provided tolerance. The offset is determined by 
        multiplying the tolerance by the sine of a random angle uniformly 
        distributed between 0 and 2π radians, simulating the effects of 
        coaxiality in all possible directions.
        
        Returns
        -------
        float
            The modified value after applying the random circular offset.
        """
        return self.specification.TOLERANCE * sin(random.uniform(0, 2 * pi))
    
    def coaxial(self) -> float:
        """Generate a random coaxiality value.
        
        This method generates a random coaxiality value by applying a 
        circular distribution to the tolerance range. The amplitude of 
        the coaxiality is determined by a normal distribution within 
        the tolerance range, and the direction is randomized using a 
        uniform distribution between 0 and 2π radians.
        
        Returns
        -------
        float
            The generated random coaxiality value.
        """
        return self.normal() * sin(random.uniform(0, 2 * pi))
    
    def perpendicular(self) -> float:
        """Generate a random perpendicularity value.
        
        This method generates a random perpendicularity value by applying 
        a circular distribution to the tolerance range. The amplitude of 
        the perpendicularity is determined by a normal distribution 
        within the tolerance range, and the direction is randomized 
        using a uniform distribution between 0 and 2π radians. In 
        principle, this is exactly the same as coaxiality.
        
        Returns
        -------
        float
            The generated random perpendicularity value.
        
        Notes
        -----
        The returned value is in the same unit as the specified 
        tolerance (usually mm). When specifying tolerances according to 
        GPS, a protractor is held against the measuring object, and the 
        maximum gap must not be larger than the specified tolerance.
        For further information, see:
        https://www.keyence.de/ss/products/measure-sys/gd-and-t/orientation-tolerance/perpendicularity.jsp
        """
        return self.normal() * sin(random.uniform(0, 2 * pi))
    
    def __call__(self) -> float:
        """Generate a random value within the specified tolerance range.
        
        This method generates a random value within the specified 
        tolerance range, clipping the value if the clip parameter is 
        set to True.
        
        Returns
        -------
        float
            The generated random value.
        """
        return getattr(self, self.dist)()
    
    def generate(self, n: int) -> NDArray:
        """Generate an array of random values.
        
        This method generates an array of random values by calling the 
        appropriate distribution method (normal or uniform) for the 
        specified number of times.
        
        Parameters
        ----------
        n : int
            The number of random values to generate.
        
        Returns
        -------
        NDArray[np.float64]
            An array of random values.
        """
        return np.array([self() for _ in range(n)])


class Binning:
    """A class for binning precise values into a specified number of 
    bins.

    This class is designed to handle scenarios where precise values are 
    available only at certain intervals with a defined tolerance. It 
    provides methods to bin these values using either a linear or 
    quantile approach.

    The binning method is determined by the `kind` parameter:
    - If `kind` is set to 'linear', bin edges are calculated based on 
      the minimum and maximum values of the provided data.
    - If `kind` is set to 'quantile', bin edges are determined by the 
      quantiles of the input data.

    The `values` method assigns each data point to the nearest bin using 
    a digitizing function. For more accurate results, the 
    `round_to_nearest` method can be used to round the bin nominal 
    values.

    Parameters
    ----------
    data : NDArray[np.float64] | Series
        The precise values to be binned and used for calculations.
    num_bins : int
        The total number of bins to create.
    kind : Literal['linear', 'quantile'], optional
        The method to calculate the bin edges, default is 'quantile'.
    distance : float | None, optional
        The distance between the bins, representing the process tolerance.
        This parameter is only relevant when `kind` is 'linear' and must 
        be specified in that case. Default is None.

    Raises
    ------
    AssertionError
        If the `distance` is not specified when `kind` is 'linear'.
    
    Examples
    --------
    Example 1: Binning with linear approach
    ```python
    import numpy as np
    import daspi as dsp
    data = np.array([1.5, 2.3, 3.7, 4.1, 5.0])
    binning = dsp.Binning(data, num_bins=3, distance=1.0, kind='linear')
    binned_values = binning.values()
    print(f'{binning.nominals=}')
    print(f'{binning.indices=}')
    print(binned_values)
    ```

    ```console
    binning.nominals=array([2.32, 3.32, 4.32])
    binning.indices=array([0, 0, 1, 2, 2])
    [2.32, 2.32, 3.32, 4.32, 4.32]
    ```

    Example 2: Binning with quantile approach
    ```python
    data = np.array([1.5, 2.3, 3.7, 4.1, 5.0])
    binning = dsp.Binning(data, num_bins=3, kind='quantile')
    binned_values = binning.values()
    print(f'{binning.nominals=}')
    print(f'{binning.indices=}')
    print(binned_values)
    ```

    ```console
    binning.nominals=array([2.3, 3.7, 4.1])
    binning.indices=array([0, 0, 1, 2, 2])
    [2.3, 2.3, 3.7, 4.1, 4.1]
    ```

    Example 3: Rounding nominals
    ```python
    data = np.array([1.5, 2.3, 3.7, 4.1, 5.0])
    binning = dsp.Binning(data, num_bins=3, kind='quantile')
    binning.round_to_nearest(nearest=5, digit=1)
    binned_values = binning.values()
    print(f'{binning.nominals=}')
    print(binned_values)
    ```

    ```console
    binning.nominals=array([2.5, 3.5, 4.0])
    [2.5, 2.5, 3.5, 4.0, 4.0]
    ```
    """
    def __init__(
            self,
            data: NDArray[np.float64] | Series,
            num_bins: int,
            kind: Literal['linear', 'quantile'] = 'quantile',
            distance: float | None = None,
            ) -> None:
        self.data = np.asarray(data)
        self.num_bins = num_bins
        self.distance = distance
        self.kind = kind
        self.edges = self._calculate_edges()
        self.indices = self._calculate_indices()
        self.nominals = self._calculate_nominals()

    def _calculate_edges(self) -> np.ndarray:
        """Calculate the bin edges based on the specified method.

        Returns
        -------
        NDArray[np.float64]
            The calculated bin edges.
        """
        if self.kind == 'linear':
            edges = np.linspace(
                np.min(self.data),
                np.max(self.data),
                self.num_bins + 1)
        else:  # kind == 'quantile'
            edges = np.quantile(
                self.data,
                np.linspace(0, 1, self.num_bins + 1))
        return edges

    def _calculate_indices(self) -> np.ndarray:
        """Convert the precise values to bin indices.

        Returns
        -------
        NDArray[np.int_]
            The binned array of integers.
        """
        binned_data = np.digitize(self.data, self.edges) - 1
        return np.clip(binned_data, 0, self.num_bins - 1)

    def _calculate_nominals(self) -> NDArray[np.float64]:
        """Calculate the bin nominals based on the specified method.

        Returns
        -------
        NDArray[np.float64]
            The bin nominals.
        """
        if self.kind == 'linear':
            assert isinstance(self.distance, (int, float)), (
                f'Specify a distance between the bins, got {self.distance}')
            base = (
                self.distance / 2
                + np.array([i * self.distance for i in range(self.num_bins)]))
            shift = np.mean(self.data) - np.mean(base)
            nominals = base + shift
        else:
            _n = self.num_bins + 1
            quantiles = [i / _n for i in range(1, _n)]
            nominals = np.array(
                [np.quantile(self.data, q) for q in quantiles])
        return nominals

    def round_to_nearest(self, nearest: int = 5, digit: int = 3) -> Self:
        """Round to the nearest multiple of `nearest`.
        
        This function rounds the input data to the nearest multiple of 
        the specified `nearest` value at the specified `digit`.

        Parameters
        ----------
        nearest : int, optional
            The multiple to round to, by default 5.
        digit : int, optional
            The number of decimal places to round to, by default 3.
        
        Returns
        -------
        Self
            The instance with the rounded nominals.
        """
        self.nominals = round_to_nearest(self.nominals, nearest, digit)
        return self

    def values(self) -> NDArray[np.float64]:
        """Get the binned values based on the calculated indices and 
        nominals.

        Returns
        -------
        NDArray[np.float64]
            The binned values.
        """
        assert hasattr(self, 'nominals'), (
            'Calculate the nominals first using the calculate_nominals method')
        return self.nominals[self.indices]


def round_to_nearest(
        x: FloatOrArray,
        nearest: int = 5,
        digit: int = 3
        ) -> FloatOrArray:
    """Round to the nearest multiple of `nearest`.
    
    This function rounds the input data to the nearest multiple of the 
    specified `nearest` value at the specified `digit`.

    Parameters
    ----------
    x : float | NDArray[np.float64] | Series
        The input data to be rounded.
    nearest : int, optional
        The multiple to round to, by default 5.
    digit : int, optional
        The number of decimal places to round to, by default 3.
    
    Returns
    -------
    float | NDArray[np.float64] | Series
        The rounded data as the same type as the input.
    """
    factor = 10**digit
    return np.round(x / nearest * factor) / factor * nearest

def inclination_displacement(
        perpendicularity: FloatOrArray,
        height: float,
        distance: float
        ) -> FloatOrArray:
    """Calculate the displacement from a distant point due to 
    perpendicularity deviation. Perpendicularity is measured in mm. To 
    calculate the angle, the height at which perpendicularity is 
    measured is required as the maximum distance in the drawing.

    $$ displacement = distance * sin(arctan(perpendicularity/height)) $$
    
    Parameters
    ----------
    perpendicularity : float | NDArray[np.float64] | Series
        The perpendicularity of the surface.
    distance : float
        The distance from the point to the surface.

    Returns
    -------
    float | NDArray[np.float64] | Series
        The displacement of the point.
    """
    displacement = distance * np.tan(np.arctan(perpendicularity/height))
    return displacement # type: ignore
